import os
import shutil
import tempfile
import unittest
import pynsive


INIT_PY = """
from .test_classes import *
"""

TEST_CLASSES_PY = """
SUCCESS = True

class PynsiveTestingClass(object):
    pass


class OtherPynsiveTestingClass(PynsiveTestingClass):
    pass
"""


class WhenLoading(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """
        This crazy setUp method for the following unit tests creates a
        temporary plugin directory and then drops a Python module and related
        testing code into it. The method then spins up a plugin context via
        a pynsive.PluginManager instance.
        """
        cls.directory = tempfile.mkdtemp()

        cls.module_path = os.path.join(cls.directory, 'pynsive_test')
        os.mkdir(cls.module_path)
        with open(os.path.join(cls.module_path, '__init__.py'), 'w') as f:
            f.write(INIT_PY)
        with open(os.path.join(cls.module_path, 'test_classes.py'), 'w') as f:
            f.write(TEST_CLASSES_PY)

        embedded_module = os.path.join(cls.module_path, 'embedded')
        os.mkdir(embedded_module)
        with open(os.path.join(embedded_module, '__init__.py'), 'w') as f:
            f.write('\n')
        with open(os.path.join(embedded_module, 'test.py'), 'w') as f:
            f.write('\n')

        cls.plugin_manager = pynsive.PluginManager()
        cls.plugin_manager.plug_into(cls.directory)

    @classmethod
    def tearDownClass(cls):
        cls.plugin_manager.destroy()
        if cls.directory:
            #shutil.rmtree(cls.directory)
            pass

    def test_plugging_into_directory(self):
        """
        When plugging into a directory using a PluginManager, the manager
        will make the new directory available for search when importing
        modules. These modules must be available for import via the
        import_module function provided by pynsive.
        """
        test_module = pynsive.import_module('pynsive_test.test_classes')
        self.assertTrue(test_module.SUCCESS)

    def test_listing_classes(self):
        classes = pynsive.list_classes('pynsive_test')
        self.assertEqual(len(classes), 2)

    def test_listing_classes_with_filter(self):
        test_module = pynsive.import_module('pynsive_test.test_classes')

        def subclasses_only(test_type):
            same = test_type is not test_module.PynsiveTestingClass
            is_subclass = issubclass(
                test_type, test_module.PynsiveTestingClass)
            return not same and is_subclass

        classes = pynsive.list_classes('pynsive_test', subclasses_only)
        self.assertEqual(len(classes), 1)

    def test_importing_missing_module(self):
        with self.assertRaises(ImportError):
            pynsive.list_modules('this.is.totally.fake.and.a.bad.name')

    def test_crawling_plugged_in_modules(self):
        found_modules = pynsive.list_modules('pynsive_test')
        self.assertTrue('pynsive_test.test_classes' in found_modules)

    def test_crawling_modules(self):
        found_modules = pynsive.list_modules('pynsive_test')
        self.assertEqual(1, len(found_modules))
        self.assertTrue('pynsive_test.test_classes' in found_modules)

    def test_crawling_modules_recursively(self):
        found_modules = pynsive.rlist_modules('pynsive_test')
        self.assertEqual(3, len(found_modules))
        self.assertTrue('pynsive_test.test_classes' in found_modules)
        self.assertTrue('pynsive_test.embedded' in found_modules)
        self.assertTrue('pynsive_test.embedded.test' in found_modules)

    def test_discovering_classes(self):
        classes = pynsive.list_classes('pynsive_test')
        self.assertEqual(2, len(classes))

    def test_recursively_listing_classes_with_filter(self):
        test_module = pynsive.import_module('pynsive_test.test_classes')

        def subclasses_only(test_type):
            same = test_type is not test_module.PynsiveTestingClass
            is_subclass = issubclass(
                test_type, test_module.PynsiveTestingClass)
            print('test class - {}'.format(test_type))
            print(same)
            print(is_subclass)
            return not same and is_subclass

        classes = pynsive.rlist_classes('pynsive_test', subclasses_only)
        self.assertEqual(len(classes), 1)



import sys

from .loader import ModuleFinder


class PluginManager(object):
    """
    The PluginManager class is a nice wrapper for tapping into sys.meta_path
    with pynsive. It also provides a hook for removing itself from
    sys.meta_path.
    """
    def __init__(self):
        self.finder = ModuleFinder()
        sys.meta_path.append(self.finder)

    def destroy(self):
        """
        Removes the meta_path hook associated with this manager.
        """
        sys.meta_path.remove(self.finder)

    def plug_into(self, *paths):
        """
        Adds all arguments passed as plugin directories to search when loading
        modules.
        """
        [self.finder.add_path(path) for path in paths]



import imp
import sys
import os

from ..common import *


class LoaderError(ImportError):
    """
    This error is thrown when the module loader encounters an exception or
    an unrecoverable state while attempting to load a dynamically located
    module.
    """
    def __init__(self, msg):
        """
        Creates a loader error.

        :param msg: the loader error message to relay upwards
        """
        self.msg = msg


class ModuleLoader(object):
    """
    As per PEP302, this module loader provides all of the necessary context
    for loading a python file and executing its contents.
    """
    def __init__(self, module_path, module_name, load_target, is_pkg):
        """
        Creates a new module loader.

        :param module_path: the path where the module resides
        :param module_name: the full name of the module to load
        :param load_target: the absolute path of the module file to load
        :param is_pkg: true if this module is a package, false otherwise
        """
        self.module_path = module_path
        self.module_name = module_name
        self.load_target = load_target
        self.is_pkg = is_pkg

    def _read_code(self):
        """
        Simple abstraction to make reading a module file easy.
        """
        fin = open(self.load_target, 'r')
        code = fin.read()
        fin.close()
        return code

    def load_module(self, module_name):
        """
        Loads a module's code and sets the module's expected hidden
        variables. For more information on these variables and what they
        are for, please see PEP302.

        :param module_name: the full name of the module to load
        """
        if module_name != self.module_name:
            raise LoaderError(
                'Requesting a module that the loader is unaware of.')

        if module_name in sys.modules:
            return sys.modules[module_name]

        code = self._read_code()
        module = imp.new_module(module_name)
        module.__file__ = self.load_target
        module.__loader__ = self

        if self.is_pkg:
            module.__path__ = [self.module_path]
            module.__package__ = module_name
        else:
            module.__package__ = module_name.rpartition('.')[0]

        sys.modules[module_name] = module
        exec(code, module.__dict__)
        return module


class ModuleFinder(object):
    """
    As per PEP302, this module loader provides all of the necessary context
    for dynamically locating python modules based. This finder searches
    directories based on paths added to its internal list of available
    search directories.
    """
    def __init__(self, paths=None):
        """
        Creates a module finder.

        :param paths: the paths to include in the search list by default
        """
        self.paths = paths if paths else list()

    def add_path(self, path):
        """
        Adds a path to search through when attempting to look up a module.

        :param path: the path the add to the list of searchable paths
        """
        if path not in self.paths:
            self.paths.append(path)

    def find_module(self, module_name, path=None):
        """
        Searches the paths for the required module.

        :param module_name: the full name of the module to find
        :param path: set to None when the module in being searched for is a
                     top-level module - otherwise this is set to
                     package.__path__ for submodules and subpackages (unused)
        """
        module_path = os.path.join(*module_name.split(MODULE_PATH_SEP))

        for search_root in self.paths:
            target_path = os.path.join(search_root, module_path)
            is_pkg = False

            # If the target references a directory, try to load it as
            # a module by referencing the __init__.py file, otherwise
            # append .py and attempt to resolve it.
            if os.path.isdir(target_path):
                target_file = os.path.join(target_path, '__init__.py')
                is_pkg = True
            else:
                target_file = '{}.py'.format(target_path)

            if os.path.exists(target_file):
                return ModuleLoader(
                    target_path, module_name, target_file, is_pkg)
        return None



import sys
import os.path
import inspect

from .common import *


def _list_classes(module, cls_filter):
    found = list()
    for name, module_obj in inspect.getmembers(module):
        if inspect.isclass(module_obj):
            append = not cls_filter or cls_filter(module_obj)
            if append:
                found.append(module_obj)
    return found


def _scan_paths_for(mname, paths):
    mpath_part = mname.replace(MODULE_PATH_SEP, os.sep)
    for mpath in paths:
        scan_target = os.path.join(mpath, mpath_part)
        is_dir = os.path.isdir(scan_target)
        is_file = os.path.isfile('{}.py'.format(scan_target))
        if is_dir or is_file:
            return mpath


def _scan_for_modules(directory, recursive=False, prefix=None):
    found = list()
    # Scan only if there's an __init__.py file
    if os.path.isfile(os.path.join(directory, MODULE_INIT_FILE)):
        for entry in os.listdir(directory):
            # Skip the init file
            if entry == MODULE_INIT_FILE:
                continue
            if entry.endswith('.py'):
                module = entry.rstrip('.py')
                if prefix:
                    found.append(MODULE_PATH_SEP.join((prefix, module)))
                else:
                    found.append(module)
            elif recursive:
                next_dir = os.path.join(directory, entry)
                if os.path.isdir(next_dir):
                    if prefix:
                        next_mod = MODULE_PATH_SEP.join((prefix, entry))
                    else:
                        next_mod = entry
                    found.append(next_mod)
                    found.extend(_scan_for_modules(next_dir, True, next_mod))
    return found


def _should_use_module_path(module):
    """
    Checks to make sure that the passed module has the correct hidden
    variables set.
    """
    return hasattr(module, PATH_ATTRUBITE) and len(module.__path__) > 0


def list_modules(mname):
    """
    Attempts to list all of the submodules under a module. This function
    works for modules located in the default path as well as extended paths
    via the sys.meta_path hooks.

    This function carries the expectation that the hidden module variable
    '__path__' has been set correctly.

    :param mname: the module name to descend into
    """
    module = import_module(mname)
    if not module:
        raise ImportError('Unable to load module {}'.format(mname))

    found = list()
    if _should_use_module_path(module):
        mpath = module.__path__[0]
    else:
        mpaths = sys.path
        mpath = _scan_paths_for(mname, mpaths)

    if mpath:
        for pmname in _scan_for_modules(mpath):
            found_mod = MODULE_PATH_SEP.join((mname, pmname))
            found.append(found_mod)
    return found


def rlist_modules(mname):
    """
    Attempts to the submodules under a module recursively. This function
    works for modules located in the default path as well as extended paths
    via the sys.meta_path hooks.

    This function carries the expectation that the hidden module variable
    '__path__' has been set correctly.

    :param mname: the module name to descend into
    """
    module = import_module(mname)
    if not module:
        raise ImportError('Unable to load module {}'.format(mname))

    found = list()
    if _should_use_module_path(module):
        mpath = module.__path__[0]
    else:
        mpaths = sys.path
        mpath = _scan_paths_for(mname, mpaths)

    if mpath:
        for pmname in _scan_for_modules(mpath, recursive=True):
            found_mod = MODULE_PATH_SEP.join((mname, pmname))
            found.append(found_mod)
    return found


def list_classes(mname, cls_filter=None):
    """
    Attempts to list all of the classes within a specified module. This
    function works for modules located in the default path as well as
    extended paths via the sys.meta_path hooks.

    If a class filter is set, it will be called with each class as its
    parameter. This filter's return value must be interpretable as a
    boolean. Results that evaluate as True will include the type in the
    list of returned classes. Results that evaluate as False will exclude
    the type in the list of returned classes.

    :param mname: of the module to descend into
    :param cls_filter: a function to call to determine what classes should be
                included.
    """
    found = list()
    module = import_module(mname)
    if inspect.ismodule(module):
        [found.append(mod) for mod in _list_classes(module, cls_filter)]
    return found


def rlist_classes(module, cls_filter=None):
    """
    Attempts to list all of the classes within a given module namespace.
    This method, unlike list_classes, will recurse into discovered
    submodules.

    If a type filter is set, it will be called with each class as its
    parameter. This filter's return value must be interpretable as a
    boolean. Results that evaluate as True will include the type in the
    list of returned classes. Results that evaluate as False will exclude
    the type in the list of returned classes.

    :param mname: of the module to descend into
    :param cls_filter: a function to call to determine what classes should be
                included.
    """
    found = list()
    mnames = list_modules(module)
    for mname in mnames:
        [found.append(c) for c in list_classes(mname, cls_filter)]
    return found
